#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_pyeasyga
----------------------------------

Tests for `pyeasyga` module.
"""

import unittest
import random
import copy
from operator import attrgetter

from pyeasyga import pyeasyga


class TestPyeasyga(unittest.TestCase):

    ''' Test Problem: Assume you can only carry three items to sell in your bag.
        Pick a combination of items that maximises your profit. The profit for
        each item is listed. '''
    def setUp(self):
        self.seed_data = [('apple', 15), ('banana', 10), ('carrot', 12),
                          ('pear', 5), ('mango', 8)]

        member_1 = [0, 1, 0, 1, 1]
        member_2 = [1, 1, 0, 0, 1]
        member_3 = [1, 1, 1, 1, 0]
        member_4 = [0, 1, 1, 1, 1]

        self.population = []
        self.population.append(member_1)
        self.population.append(member_2)
        self.population.append(member_3)
        self.population.append(member_4)

        self.ga = pyeasyga.GeneticAlgorithm(self.seed_data)
        self.ga.population_size = 10
        self.ga.generations = 10

        self.ga.fitness_function = lambda member, data: sum(
            [profit for (selected, (fruit, profit)) in
             zip(member, data) if selected and
             member.count(1) == 3])

        def mutate(individual):
            mutate_index = random.randrange(len(individual))
            individual[mutate_index] = (0, 1)[individual[mutate_index] == 0]

        self.ga.selection_function = self.ga.tournament_selection

    def test_genetic_algorithm_initialisation_1(self):
        ''' Test default initialisation '''
        ga_1 = pyeasyga.GeneticAlgorithm(self.seed_data)

        assert ga_1.population_size == 100
        assert ga_1.generations == 300
        assert ga_1.elitism is True
        assert ga_1.crossover_probability == 0.8
        assert ga_1.mutation_probability == 0.2

    def test_genetic_algorithm_initialisation_2(self):
        ''' Test initialisation with specific values '''
        ga_2 = pyeasyga.GeneticAlgorithm(
            self.seed_data, population_size=200, generations=500,
            crossover_probability=0.9, mutation_probability=0.05)
        ga_2.elitism = False
        ga_2.maximise_fitness = False
        fruit, profit = ga_2.seed_data[1]

        assert len(ga_2.seed_data) == 5
        assert (fruit, profit) == ('banana', 10)
        assert ga_2.population_size == 200
        assert ga_2.generations == 500
        assert ga_2.elitism is False
        assert ga_2.maximise_fitness is False
        assert ga_2.crossover_probability == 0.9
        assert ga_2.mutation_probability == 0.05

    def test_chromosome_initialisation_1(self):
        chromosome = pyeasyga.Chromosome(['a', 'b', 'c'])

        assert chromosome.genes == ['a', 'b', 'c']
        assert chromosome.fitness == 0
        assert str(chromosome) == "(0, ['a', 'b', 'c'])"

    def test_chromosome_initialisation_2(self):
        chromosome = pyeasyga.Chromosome(['d', 'e', 'f'])
        chromosome.fitness = 20

        assert chromosome.genes == ['d', 'e', 'f']
        assert chromosome.fitness == 20
        assert str(chromosome) == "(20, ['d', 'e', 'f'])"

    def test_fitness_function(self):
        func = self.ga.fitness_function
        data = self.ga.seed_data

        assert func([0, 1, 0, 1, 1], data) == 23
        assert func([1, 1, 0, 0, 1], data) == 33
        assert func([1, 1, 1, 1, 0], data) == 0
        assert func([0, 1, 1, 1, 1], data) == 0

    def test_crossover_function(self):
        parent_1 = [1, 1, 1, 1, 1]
        parent_2 = [0, 0, 0, 0, 0]
        child_1, child_2 = self.ga.crossover_function(parent_1, parent_2)

        assert child_1 not in [[0, 0, 0, 0, 0], [1, 1, 1, 1, 1]]
        assert child_2 not in [[0, 0, 0, 0, 0], [1, 1, 1, 1, 1]]

    def test_mutate_function(self):
        ''' Check that the individual is not the same after mutation
            Check that only one bit in the individual is changed '''
        individual = [0, 1, 1, 0, 1]
        self.ga.mutate_function(individual)

        res = [x == y for (x, y) in zip(individual, [0, 1, 1, 0, 1])]

        assert individual != [0, 1, 1, 0, 1]
        assert res.count(False) == 1

    def test_selection_function_1(self):
        ''' Test tournament selection '''
        population = []

        member_1 = pyeasyga.Chromosome([0, 1, 0, 1, 1])
        member_2 = pyeasyga.Chromosome([1, 1, 0, 0, 1])
        member_3 = pyeasyga.Chromosome([1, 1, 1, 1, 0])
        member_4 = pyeasyga.Chromosome([0, 1, 1, 1, 1])

        population.append(member_1)
        population.append(member_2)
        population.append(member_3)
        population.append(member_4)

        self.ga.calculate_population_fitness(
            self.ga.seed_data, population, self.ga.fitness_function)
        self.ga.tournament_size = 4
        self.ga.selection_function = self.ga.tournament_selection

        individual = self.ga.selection_function(population)

        assert individual.genes == [1, 1, 0, 0, 1]
        assert individual.fitness == 33
        assert len(individual.genes) == 5
        assert individual in population

    def test_selection_function_2(self):
        ''' Test tournament selection '''
        population = []

        member_1 = pyeasyga.Chromosome([0, 1, 0, 1, 1])
        member_2 = pyeasyga.Chromosome([1, 1, 0, 0, 1])
        member_3 = pyeasyga.Chromosome([1, 1, 1, 1, 0])
        member_4 = pyeasyga.Chromosome([0, 1, 1, 1, 1])

        population.append(member_1)
        population.append(member_2)
        population.append(member_3)
        population.append(member_4)

        self.ga.calculate_population_fitness(
            self.ga.seed_data, population, self.ga.fitness_function)
        self.ga.tournament_size = 0
        self.ga.selection_function = self.ga.tournament_selection

        individual = self.ga.selection_function(population)

        assert len(individual.genes) == 5
        assert individual in population

    def test_selection_function_3(self):
        ''' Test random selection '''
        self.ga.selection_function = self.ga.random_selection
        individual = self.ga.selection_function(self.population)

        assert len(individual) == 5
        assert individual in self.population

    def test_create_individual(self):
        data = self.ga.seed_data
        individual = self.ga.create_individual(data)

        assert len(individual) == 5
        assert all([ind in (0, 1) for ind in individual if ind in (0, 1)])

    def test_create_initial_population(self):
        pop_size = self.ga.population_size
        func = self.ga.create_individual
        data = self.ga.seed_data

        initial_population = self.ga.create_initial_population(
            data, pop_size, func)

        assert len(initial_population) == pop_size
        assert isinstance(initial_population[0],
                          type(pyeasyga.Chromosome([1])))
        assert sum([member.fitness for member in initial_population]) == 0

    def test_calculate_population_fitness(self):
        data = self.ga.seed_data
        pop_size = self.ga.population_size
        create_func = self.ga.create_individual

        population = self.ga.create_initial_population(
            data, pop_size, create_func)

        fitness_func = self.ga.fitness_function
        self.ga.calculate_population_fitness(data, population, fitness_func)

        assert sum([member.fitness for member in population]) > 0

    def test_rank_population(self):
        data = self.seed_data

        pop_size = self.ga.population_size
        create_func = self.ga.create_individual
        fitness_func = self.ga.fitness_function

        population = self.ga.create_initial_population(
            data, pop_size, create_func)
        self.ga.calculate_population_fitness(data, population, fitness_func)

        new_population = copy.deepcopy(population)
        new_population.sort(key=attrgetter('fitness'), reverse=True)

        self.ga.rank_population(population, True)

        assert population[0].fitness == new_population[0].fitness
        assert population[1].fitness == new_population[1].fitness

    def test_create_new_population(self):
        data = self.seed_data

        pop_size = self.ga.population_size
        create_func = self.ga.create_individual
        fitness_func = self.ga.fitness_function

        population = self.ga.create_initial_population(
            data, pop_size, create_func)
        self.ga.calculate_population_fitness(data, population, fitness_func)
        self.ga.rank_population(population, True)

        crossover = self.ga.crossover_function
        prob_crossover = self.ga.crossover_probability
        mutate = self.ga.mutate_function
        prob_mutate = self.ga.mutation_probability
        selection = self.ga.selection_function
        elitism = True

        new_population = self.ga.create_new_population(
            population, crossover, prob_crossover, mutate, prob_mutate,
            selection, elitism)

        assert len(new_population) == self.ga.population_size
        assert new_population[0].fitness == population[0].fitness
        assert new_population[0].genes == population[0].genes

    def test_create_first_generation(self):
        data = self.ga.seed_data
        pop_size = self.ga.population_size
        create_func = self.ga.create_individual
        fitness_func = self.ga.fitness_function
        max_fitness = self.ga.maximise_fitness

        population = self.ga.create_first_generation(
            data, pop_size, create_func, fitness_func, max_fitness)

        assert len(population) == pop_size
        assert isinstance(population[0], type(pyeasyga.Chromosome([1])))

    def test_create_next_generation(self):
        data = self.ga.seed_data
        pop_size = self.ga.population_size
        create_func = self.ga.create_individual
        fitness_func = self.ga.fitness_function
        selection = self.ga.selection_function
        crossover = self.ga.crossover_function
        prob_crossover = self.ga.crossover_probability
        mutate = self.ga.mutate_function
        prob_mutate = self.ga.mutation_probability
        elitism = self.ga.elitism
        max_fitness = self.ga.maximise_fitness

        population = self.ga.create_first_generation(
            data, pop_size, create_func, fitness_func, max_fitness)

        next_gen = self.ga.create_next_generation(
            data, population, fitness_func, selection, crossover,
            prob_crossover, mutate, prob_mutate, elitism, max_fitness)

        assert len(next_gen) == pop_size
        assert isinstance(population[0], type(pyeasyga.Chromosome([1])))

    def test_run(self):
        self.ga.run()
        current_gen = self.ga.current_generation
        last_generation = self.ga.last_generation()

        assert len(current_gen) == self.ga.population_size
        assert isinstance(current_gen[0], type(pyeasyga.Chromosome([1])))
        assert isinstance(last_generation.next(), type((1, [1, 1, 1, 1, 1])))
        assert len(last_generation.next()) == 2
        assert len(last_generation.next()[1]) == 5

    def test_best_individual(self):
        data = self.seed_data
        pop_size = self.ga.population_size
        create_func = self.ga.create_individual
        fitness_func = self.ga.fitness_function
        max_fitness = self.ga.maximise_fitness

        self.ga.current_generation = self.ga.create_first_generation(
            data, pop_size, create_func, fitness_func, max_fitness)

        best_fitness, best_genes = self.ga.best_individual()

        assert best_fitness == self.ga.current_generation[0].fitness
        assert best_genes == self.ga.current_generation[0].genes

    def test_last_generation(self):
        data = self.ga.seed_data
        pop_size = self.ga.population_size
        create_func = self.ga.create_individual
        fitness_func = self.ga.fitness_function
        max_fitness = self.ga.maximise_fitness

        self.ga.current_generation = self.ga.create_first_generation(
            data, pop_size, create_func, fitness_func, max_fitness)
        last_generation = self.ga.last_generation()

        assert isinstance(last_generation.next(), type((1, [1, 1, 1, 1, 1])))
        assert len(last_generation.next()) == 2
        assert len(last_generation.next()[1]) == 5

    def tearDown(self):
        pass

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""pyeasyga

A simple and easy-to-use genetic algorithm implementation library in Python.

For a bit array solution representation, simply instantiate the
GeneticAlgorithm class with input data, define and supply a fitness function,
run the Genetic Algorithm, and retrieve the solution!

Other solution representations will require setting some more attributes.

"""

__author__ = 'Ayodeji Remi-Omosowon'
__email__ = 'remiomosowon@gmail.com'
__version__ = '0.2.0'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import re
import codecs

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(here, *parts), 'r').read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                version_file, re.M)
    if version_match:
        return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

requirements = [
]

test_requirements = ['tox']

setup(
    name='pyeasyga',
    version=find_version('pyeasyga', '__init__.py'),
    description='A simple and easy-to-use Genetic Algorithm implementation library in Python',
    long_description=readme + '\n\n' + history,
    author='Ayodeji Remi-Omosowon',
    author_email='remiomosowon@gmail.com',
    url='https://github.com/remiomosowon/pyeasyga',
    packages=[
        'pyeasyga',
    ],
    package_dir={'pyeasyga':
                 'pyeasyga'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='pyeasyga',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-
"""
    pyeasyga module

"""

import random
import copy
from operator import attrgetter


class GeneticAlgorithm(object):
    """ Genetic Algorithm class.

    This is the main class that controls the functionality of the Genetic
    Algorithm.

    >>> # Select only two items from the list and maximise profit
    >>> from pyeasyga.pyeasyga import GeneticAlgorithm
    >>> input_data = [('pear', 50), ('apple', 35), ('banana', 40)]
    >>> easyga = GeneticAlgorithm(input_data)
    >>> def fitness (member, data):
    >>>     return sum([profit for (selected, (fruit, profit)) in
    >>>                 zip(member, data) if selected and
    >>>                 member.count(1) == 2])
    >>> easyga.fitness_function = fitness
    >>> easyga.run()
    >>> print easyga.best_individual()

    """

    def __init__(self,
                 seed_data,
                 population_size=100,
                 generations=300,
                 crossover_probability=0.8,
                 mutation_probability=0.2):
        """ Instantiate the Genetic Algorithm.

        :param seed_data: input data to the Genetic Algorithm
        :type seed_data: list of objects
        :param int population_size: size of population
        :param int generations: number of generations to evolve
        :param float crossover_probability: probability of crossover operation
        :param float mutation_probability: probability of mutation operation
        :returns: None

        """

        self.seed_data = seed_data
        self.population_size = population_size
        self.generations = generations
        self.crossover_probability = crossover_probability
        self.mutation_probability = mutation_probability

        self.current_generation = []
        self.elitism = True
        self.maximise_fitness = True

        def create_individual(seed_data):
            """ Create a candidate solution representation.

            e.g. for a bit array representation:

            >>> return [random.randint(0, 1) for _ in xrange(len(data))]

            :param seed_data: input data to the Genetic Algorithm
            :type seed_data: list of objects
            :returns: candidate solution representation as a list

            """
            return [random.randint(0, 1) for _ in xrange(len(seed_data))]

        def crossover(parent_1, parent_2):
            """ Crossover (mate) two parents to produce two children

            :param parent_1: candidate solution representation (list)
            :param parent_2: candidate solution representation (list)
            :returns: tuple containing two children

            """
            index = random.randrange(1, len(parent_1))
            child_1 = parent_1[:index] + parent_2[index:]
            child_2 = parent_2[:index] + parent_1[index:]
            return child_1, child_2

        def mutate(individual):
            mutate_index = random.randrange(len(individual))
            individual[mutate_index] = (0, 1)[individual[mutate_index] == 0]

        def random_selection(population):
            return random.choice(population)

        def tournament_selection(population):
            if self.tournament_size == 0:
                return random.choice(population)
            members = random.sample(population, self.tournament_size)
            members.sort(
                key=attrgetter('fitness'), reverse=self.maximise_fitness)
            return members[0]

        self.fitness_function = None
        self.tournament_selection = tournament_selection
        self.tournament_size = self.population_size / 10
        self.random_selection = random_selection
        self.create_individual = create_individual
        self.crossover_function = crossover
        self.mutate_function = mutate
        self.selection_function = self.tournament_selection

    def create_initial_population(
            self, seed_data, population_size, create_individual):
        """ Create members of the first population randomly.

        :param seed_data: input data to the Genetic Algorithm
        :type seed_data: list of objects
        :param int population_size: size of population
        :param create_individual: function that creates candidate solutions
        :type create_individual: :func:
        :returns: initial population as a list of candidate solutions

        """
        initial_population = []
        for _ in xrange(population_size):
            genes = create_individual(seed_data)
            individual = Chromosome(genes)
            initial_population.append(individual)
        return initial_population

    def calculate_population_fitness(self, data, population, fitness_function):
        for individual in population:
            individual.fitness = fitness_function(individual.genes, data)

    def rank_population(self, population, maximise_fitness=True):
        population.sort(key=attrgetter('fitness'), reverse=maximise_fitness)

    def create_new_population(
            self, population, crossover, prob_crossover, mutate, prob_mutate,
            selection, elitism):
        new_population = []
        best = copy.deepcopy(population[0])

        while len(new_population) < len(population):
            parent_1 = copy.deepcopy(selection(population))
            parent_2 = copy.deepcopy(selection(population))

            can_crossover = random.random() < prob_crossover
            can_mutate = random.random() < prob_mutate

            child_1, child_2 = parent_1, parent_2
            child_1.fitness, child_2.fitness = 0, 0

            if can_crossover:
                child_1.genes, child_2.genes = crossover(parent_1.genes,
                                                         parent_2.genes)

            if can_mutate:
                mutate(child_1.genes)
                mutate(child_2.genes)

            new_population.append(child_1)
            if len(new_population) < len(population):
                new_population.append(child_2)

        if elitism:
            new_population[0] = best

        return new_population

    def create_first_generation(
            self, data, population_size, create_individual, fitness_function,
            maximise_fitness=True):
        initial_population = self.create_initial_population(
            data, population_size, create_individual)
        self.calculate_population_fitness(
            data, initial_population, fitness_function)
        self.rank_population(initial_population, maximise_fitness)
        return initial_population

    def create_next_generation(
            self, data, population, fitness_function,
            selection_function, crossover_function, prob_crossover,
            mutate_function, prob_mutate, elitism, maximise_fitness=True):
        new_population = self.create_new_population(
            population, crossover_function, prob_crossover, mutate_function,
            prob_mutate, selection_function, elitism)
        self.calculate_population_fitness(
            data, new_population, fitness_function)
        self.rank_population(new_population, maximise_fitness)
        return new_population

    def run(self):
        data = self.seed_data
        pop_size = self.population_size
        generations = self.generations
        create_func = self.create_individual
        fitness_func = self.fitness_function
        selection = self.selection_function
        crossover = self.crossover_function
        prob_crossover = self.crossover_probability
        mutate = self.mutate_function
        prob_mutate = self.mutation_probability
        elitism = self.elitism
        max_fitness = self.maximise_fitness

        self.current_generation = self.create_first_generation(
            data, pop_size, create_func, fitness_func, max_fitness)

        for _ in xrange(1, generations):
            next_generation = self.create_next_generation(
                data, self.current_generation, fitness_func, selection,
                crossover, prob_crossover, mutate, prob_mutate, elitism,
                max_fitness)
            self.current_generation = next_generation

    def best_individual(self):
        best = self.current_generation[0]
        return (best.fitness, best.genes)

    def last_generation(self):
        return ((member.fitness, member.genes) for member
                in self.current_generation)


class Chromosome(object):
    def __init__(self, genes):
        self.genes = genes
        self.fitness = 0

    def __repr__(self):
        return repr((self.fitness, self.genes))
